@@ -27,9 +27,10 @@ from lightly.openapi_generated.swagger_client import TagData, TagArithmeticsRequ
 
 
 def _download_cli(cfg, is_cli_call=True):
-    tag_name = cfg['tag_name']
-    dataset_id = cfg['dataset_id']
-    token = cfg['token']
+
+    tag_name = str(cfg['tag_name'])
+    dataset_id = str(cfg['dataset_id'])
+    token = str(cfg['token'])
 
     if not tag_name or not token or not dataset_id:
         print_as_warning('Please specify all of the parameters tag_name, token and dataset_id')
@@ -41,37 +42,16 @@ def _download_cli(cfg, is_cli_call=True):
     )
 
     # get tag id
-    tag_name_id_dict = dict([tag.name, tag.id] for tag in api_workflow_client._get_all_tags())
-    tag_id = tag_name_id_dict.get(tag_name, None)
-    if tag_id is None:
-        warnings.warn(f'The specified tag {tag_name} does not exist.')
-        return
-
-    # get tag data
-    tag_data: TagData = api_workflow_client.tags_api.get_tag_by_tag_id(
-        dataset_id=dataset_id, tag_id=tag_id
+    tag_data = api_workflow_client.get_tag_by_name(tag_name)
+    filenames_tag = api_workflow_client.get_filenames_in_tag(
+        tag_data,
+        exclude_parent_tag=cfg['exclude_parent_tag'],
     )
 
-    if cfg["exclude_parent_tag"]:
-        parent_tag_id = tag_data.prev_tag_id
-        tag_arithmetics_request = TagArithmeticsRequest(
-            tag_id1=tag_data.id,
-            tag_id2=parent_tag_id,
-            operation=TagArithmeticsOperation.DIFFERENCE)
-        bit_mask_response: TagBitMaskResponse \
-            = api_workflow_client.tags_api.perform_tag_arithmetics(body=tag_arithmetics_request, dataset_id=dataset_id)
-        bit_mask_data = bit_mask_response.bit_mask_data
-    else:
-        bit_mask_data = tag_data.bit_mask_data
-
-    # get samples
-    chosen_samples_ids = BitMask.from_hex(bit_mask_data).to_indices()
-    samples = [api_workflow_client.filenames_on_server[i] for i in chosen_samples_ids]
-
     # store sample names in a .txt file
-    filename = cfg['tag_name'] + '.txt'
+    filename = tag_name + '.txt'
     with open(filename, 'w') as f:
-        for item in samples:
+        for item in filenames_tag:
             f.write("%s\n" % item)
 
     filepath = os.path.join(os.getcwd(), filename)
@@ -92,7 +72,7 @@ def _download_cli(cfg, is_cli_call=True):
         dataset = data.LightlyDataset(input_dir=input_dir)
 
         # dump the dataset in the output directory
-        dataset.dump(output_dir, samples)
+        dataset.dump(output_dir, filenames_tag)
 
 
 @hydra.main(config_path='config', config_name='config')
